<?php

namespace InSegment\ApiCore\Services;

use InSegment\ApiCore\Models\SliceRemap;
use InSegment\ApiCore\Models\SliceOperation;
use InSegment\ApiCore\Models\GenerationsMatch;

use InSegment\ApiCore\Services\SliceMerger;
use InSegment\ApiCore\Exceptions\ApiTransactorException;

use Illuminate\Support\Facades\DB;

abstract class SliceOperationBase
{
    /**
     * Tables which are copied
     *
     * @var string[]
     */
    protected $copyTables = [];
    
    /**
     * Tables which are targeted as id sources
     * 
     * @var string[]
     */
    protected $targetTables = [];
    
    /**
     * Slice Operation
     *
     * @var \InSegment\ApiCore\Models\SliceOperation
     */
    protected $slice;
    
    /**
     * Slice Merger
     *
     * @var InSegment\ApiCore\Services\SliceMerger 
     */
    protected $sliceMerger;
    
    /**
     * Tables to the keys which are mentioned in some tables and thus may have updated key
     *
     * @var array [
     *     string $tableName => string|string[] $keys,
     *     ...
     * ]
     */
    protected $uuidableMentions = [];
    
    /**
     * Tables to the dependent tables with their dependent keys to be related to that table
     *
     * @var array [
     *     string $tableName => [
     *          string $dependentTable => string[] $dependentKeys
     *     ],
     *     ...
     * ]
     */
    protected $uuidableMergeMentions = [];
    
    /**
     * Tables to temporary tables
     *
     * @var array [
     *     string $tableName => [
     *         string $type => string $tempTableOfType,
     *         ...
     *     ],
     *     ...
     * ] 
     */
    protected $tempTables = [];
    
    /**
     * Counters to check estimates against those for affections when inserting records
     *
     * @var array [
     *     'new' => [
     *          string $table => int $amount,
     *          ...
     *     ],
     *     'written' => [
     *          string $table => int $amount,
     *          ...
     *     ]
     * ] 
     */
    protected $estimateCounters = [];
    
    /**
     * Class name of auto key remapper
     *
     * @var string 
     */
    protected $autoKeyRemapClass = SliceRemap::class;
    
    /**
     * Merge slice tables into real ones
     * 
     * @param callable $onMerge
     * @param callable $postProcessing
     * @return bool
     */
    public function finishCopy(callable $onMerge, callable $postProcessing)
    {
        $autoKeyRemapClass = $this->autoKeyRemapClass;
        $mergerOptions = [
            'tempTables' => &$this->tempTables,
            'uuidableMentions' => $this->uuidableMergeMentions,
            'postProcessing' => $postProcessing,
            'supplementaries' => [
                'autoKeyRemap' => $autoKeyRemapClass,
                'generationsMatch' => GenerationsMatch::class
            ]
        ];

        $this->slice->setCounters($this->estimateCounters);
        try {
            DB::transaction(function () use (&$mergerOptions, $onMerge) {
                $onMerge($this->slice);
                $this->sliceMerger->mergeTempTables($mergerOptions);
            });
        } catch (ApiTransactorException $exception) {
            $this->slice->status = SliceOperation::STATUS_ABORT;
            $this->slice->save();
            throw $exception;
        }

        foreach ($this->tempTables as $tempTableDef) {
            $this->sliceMerger->dropTempTable($tempTableDef);
        }

        $autoKeyRemapClass::disband($this->slice->getUID());
        $this->slice->status = SliceOperation::STATUS_DONE;
        return $this->slice->save();
    }
    
    /**
     * Get name of temporary table for model class
     * 
     * @param string $modelClass
     * @return string|null
     */
    public function getTempInsertTable(string $modelClass)
    {
        $modelExample = (new $modelClass);
        $modelTable = $modelExample->getTable();
        return $this->tempTables[$modelTable][SliceMerger::TEMP_TABLE_TYPE_INSERT] ?? false;
    }
    
    /**
     * Get name of temporary updates table for model class
     * 
     * @param string $modelClass
     * @return string|null
     */
    public function getTempUpdateTable(string $modelClass)
    {
        $modelExample = (new $modelClass);
        $modelTable = $modelExample->getTable();
        return $this->tempTables[$modelTable][SliceMerger::TEMP_TABLE_TYPE_UPDATE] ?? false;
    }
    
    /**
     * Get name of temporary changes table for model class
     * 
     * @param string $modelClass
     * @return string|null
     */
    public function getTempChangesTable(string $modelClass)
    {
        $modelExample = (new $modelClass);
        $modelTable = $modelExample->getTable();
        return $this->tempTables[$modelTable][SliceMerger::TEMP_TABLE_TYPE_CHANGES] ?? false;
    }
    
}
